#
# Copyright (c) nexB Inc. and others. All rights reserved.
# Copyright (c) 2018 Peter Odding
# Author: Peter Odding <peter@peterodding.com>
# URL: https://github.com/xolox/python-deb-pkg-tools
# SPDX-License-Identifier: Apache-2.0 AND MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.

import re

from attr import asdict
from attr import attrs
from attr import attrib

from debian_inspector import version as dversion

"""
Parse and evaluate Debian package relationship aka. dependencies.

This module provides functions to parse and evaluate Debian package relationship
declarations as defined in `chapter 7` of the Debian policy manual.
http://www.debian.org/doc/debian-policy/ch-relationships.html#s-depsyntax
"""

# Define a compiled regular expression pattern that we will use to match
# package relationship expressions consisting of a package name followed by
# optional version and architecture restrictions.
parse_package_relationship_expression = re.compile(r'''
    # Capture all leading characters up to (but not including)
    # the first parenthesis, bracket or space.
    (?P<name> [^\(\[ ]+ )
    # Ignore any whitespace.
    \s*
    # Optionally capture version restriction inside parentheses.
    ( \( (?P<version> [^)]+ ) \) )?
    # Ignore any whitespace.
    \s*
    # Optionally capture architecture restriction inside brackets.
    ( \[ (?P<architectures> [^\]]+ ) \] )?
''', re.VERBOSE).match

ARCHITECTURE_RESTRICTIONS_MESSAGE = "Architecture constraint is not implemented."


def parse_depends(relationships):
    """
    Return an AndRelationships from a Debian package relationship declaration
    line. Raise ValueError when parsing fails.

    `relationships` is a string containing one or more comma separated package
    relationships or a list of strings with package relationships.

    The input has a form such as ``python (>= 2.6), python (<< 3)``, i.e. a
    comma-separated list of expressions.

    Each expression is a package name/versions/arch constraint or a pipe-
    separated list of alternative.
    """

    if isinstance(relationships, str):
        relationships = (r.strip() for r in relationships.split(',') if r.strip())

    return AndRelationships.from_relationships(*(map(parse_alternatives, relationships)))


def parse_alternatives(expression):
    """
    Return a Relationship for an expression string that contains one or more
    alternative relationships. Raise ValueError when parsing fails.

    An expression with alternative is something such as as ``python2.6 |
    python2.7.``, i.e. a list of relationship expressions separated by ``|``
    tokens.

    Each pipe-separated sub-expression is parsed with `parse_relationship()`
    """
    if '|' in expression:
        alternatives = (a.strip() for a in expression.split('|') if a.strip())
        alternatives = (parse_relationship(a) for a in alternatives)
        return OrRelationships.from_relationships(*alternatives)
    else:
        return parse_relationship(expression)


# split on operators
split_on_ops = re.compile('([<>=]+)').split


def parse_relationship(expression):
    """
    Parse an expression containing a package name and optional version and
    architecture constraints.

    :param expression: A relationship expression (a string).
    :returns: A :class:`Relationship` object.
    :raises: :exc:`~exceptions.ValueError` when parsing fails.

    This function parses relationship expressions containing a package name and
    (optionally) a version relation of the form ``python (>= 2.6)`` and/or an
    architecture restriction (refer to the Debian policy manual's documentation
    on the syntax of relationship fields for details).
    https://www.debian.org/doc/debian-policy/ch-relationships.html
    """
    try:
        pre = parse_package_relationship_expression(expression)
        name = pre.group('name')
        version = pre.group('version')
    except AttributeError:
        print('gd:', pre.groupdict())
        print('this:', repr(expression))
        raise

    # Split the architecture restrictions into a tuple of strings.
    architectures = tuple((pre.group('architectures') or '').split())

    if name and not version:
        # A package name (and optional architecture restrictions) without
        # version relation.
        return Relationship(name=name, architectures=architectures)

    else:
        # A package name (and optional architecture restrictions) followed by a
        # relationship to specific version(s) of the package.
        tokens = [t.strip() for t in split_on_ops(version) if t and t.strip()]
        if len(tokens) != 2:
            # Encountered something unexpected!
            raise ValueError(
                'Corrupt package relationship expression: Splitting operator '
                'from version resulted in more than two tokens! '
                '(expression: {e}, tokens: {t})'.format(e=expression, t=tokens)
            )
        return VersionedRelationship(
            name=name,
            architectures=architectures,
            operator=tokens[0],
            version=tokens[1])


@attrs
class AbstractRelationship(object):

    @property
    def names(self):
        """
        A set of package names (strings) in the relationship.
        """
        raise NotImplementedError

    def matches(self, name, version=None, architecture=None):
        """
        Check if the relationship matches a given package name and version.

        Return True, False or None:
          - True if the name and version match,
          - False if the name matches but not the version
          - None if the name does not match
        """
        raise NotImplementedError


@attrs
class Relationship(AbstractRelationship):
    """
    A simple package relationship referring only to the name of a package.
    """

    name = attrib()
    architectures = attrib(default=tuple())

    @property
    def names(self):
        return set([self.name])

    def matches(self, name, version=None, architecture=None):
        """
        Return True if the relationship matches a given package name ignoring
        version or None.
        """
        if self.name == name:
            # TODO: support architectures
            if self.architectures:
                raise NotImplementedError(ARCHITECTURE_RESTRICTIONS_MESSAGE)
            return True
        else:
            return None

    def __str__(self, *args, **kwargs):
        if not self.architectures:
            return self.name

        return '{name} {arches}'.format(
            name=self.name,
            arches='[{}]'.format(' '.join(self.architectures))
        )

    def to_dict(self):
        return asdict(self)


@attrs
class VersionedRelationship(Relationship):
    """
    A conditional package relationship that refers to a package and certain
    versions of that package.
    """

    name = attrib()
    operator = attrib()
    version = attrib()
    architectures = attrib(default=tuple())

    def matches(self, name, version=None, architecture=None):
        """
        Check if the relationship matches a given package name and version.
        """
        if self.name == name:
            if version:
                if self.architectures:
                    raise NotImplementedError(ARCHITECTURE_RESTRICTIONS_MESSAGE)
                # TODO: check if this is the correct order
                return dversion.eval_constraint(version, self.operator, self.version)
            else:
                return False
        else:
            return None

    def __str__(self, *args, **kwargs):
        s = f'{self.name} ({self.operator} {self.version})'
        if self.architectures:
            s += ' [{}]'.format(' '.join(self.architectures))
        return s


@attrs
class MultipleRelationship(AbstractRelationship):

    relationships = attrib(default=tuple())

    @classmethod
    def from_relationships(cls, *relationships):
        return cls(relationships=tuple(relationships))

    @property
    def names(self):
        """
        Get the name(s) of the packages in the relationship set.

        :returns: A set of package names (strings).
        """
        names = set()
        for relationship in self.relationships:
            names |= relationship.names
        return names

    def __iter__(self):
        return iter(self.relationships)


class OrRelationships(MultipleRelationship):
    """
    A multi relationship where one of the relationships must be satisfied.
    """

    def matches(self, name, version=None, architecture=None):
        """
        Return True if at least one of the relationships match a given package
        name and version strings.

        Return False if at least one of the relationships name are matched
        without the version or None otherwise.
        """
        matches = None
        for alternative in self.relationships:
            alternative_matches = alternative.matches(name, version, architecture)
            if alternative_matches is True:
                return True
            elif alternative_matches is False:
                # Keep looking for a match but return False if we don't find one.
                matches = False
        return matches

    def __str__(self, *args, **kwargs):
        return ' | '.join(str(r) for r in self.relationships)


@attrs
class AndRelationships(MultipleRelationship):
    """
    A multi relationship where all relationships must be satisfied.
    """

    def matches(self, name, version=None, architecture=None):
        """
        Return True if all the the relationships match a given package name and
        version strings.

        Return False if at least one relationship matches to False or None
        otherwise.
        """
        results = (r.matches(name, version) for r in self.relationships)
        matches = [r for r in results if r is not None]
        if matches:
            return all(matches)
        else:
            return None

    def __str__(self, *args, **kwargs):
        return ', '.join(str(r) for r in self.relationships)




#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#


from unittest import TestCase

from debian_inspector import unsign


class Testunsign(TestCase):

    def test_remove_signature_and_is_signed_corner_cases(self):
        assert not unsign.is_signed(None)
        assert unsign.remove_signature(None) == None

        assert not unsign.is_signed('')
        assert unsign.remove_signature('') == ''

        assert not unsign.is_signed('\n')
        assert unsign.remove_signature('\n') == '\n'

        assert not unsign.is_signed('sometext\n')
        assert unsign.remove_signature('sometext\n') == 'sometext\n'

    def test_remove_signature_and_is_signed_not_signed(self):
        assert not unsign.is_signed(PLAIN)
        assert unsign.remove_signature(PLAIN) == PLAIN

    def test_remove_signature_and_is_signed_empty_signed(self):
        assert unsign.is_signed(EMPTY)
        assert unsign.remove_signature(EMPTY) == '\n'

    def test_remove_signature_and_is_signed_signed(self):
        assert unsign.is_signed(SIGNED)
        assert unsign.remove_signature(SIGNED) == EXPECTED_SIGNED

    def test_remove_signature_and_is_signed_compact(self):
        assert unsign.is_signed(COMPACT)
        assert unsign.remove_signature(COMPACT) == EXPECTED_COMPACT

    def test_is_signed1(self):
        text = '''
        -----BEGIN PGP SIGNED MESSAGE-----
        -----END PGP SIGNATURE-----
        '''
        assert unsign.is_signed(text)

    def test_is_signed2(self):
        text = '''-----BEGIN PGP SIGNED MESSAGE-----
        -----END PGP SIGNATURE-----'''
        assert unsign.is_signed(text)

    def test_is_signed3(self):
        text = '''-----BEGIN PGP SIGNED MESSAGE-----'''
        assert not unsign.is_signed(text)

    def test_is_signed4(self):
        text = '''-----END PGP SIGNATURE-----'''
        assert not unsign.is_signed(text)


PLAIN = '''Format: 3.0 (quilt)
Source: zlib
Binary: zlib1g, zlib1g-dev, zlib1g-dbg, zlib1g-udeb, lib64z1, lib64z1-dev, lib32z1, lib32z1-dev, libn32z1, libn32z1-dev
Architecture: any
Version: 1:1.2.11.dfsg-1
Maintainer: Mark Brown <broonie@debian.org>
Homepage: http://zlib.net/
Standards-Version: 3.9.8
'''

SIGNED = '''-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA512

Format: 3.0 (quilt)
Source: zlib
Binary: zlib1g, zlib1g-dev, zlib1g-dbg, zlib1g-udeb, lib64z1, lib64z1-dev, lib32z1, lib32z1-dev, libn32z1, libn32z1-dev
Architecture: any
Version: 1:1.2.11.dfsg-1
Maintainer: Mark Brown <broonie@debian.org>
Homepage: http://zlib.net/
Standards-Version: 3.9.8
Build-Depends: debhelper (>= 8.1.3~), binutils (>= 2.18.1~cvs20080103-2) [mips mipsel], gcc-multilib [amd64 i386 kfreebsd-amd64 mips mipsel powerpc ppc64 s390 sparc s390x] <!stage1>, dpkg-dev (>= 1.16.1)
Package-List:
 lib32z1 deb libs optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib32z1-dev deb libdevel optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib64z1 deb libs optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 lib64z1-dev deb libdevel optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 libn32z1 deb libs optional arch=mips,mipsel profile=!stage1
 libn32z1-dev deb libdevel optional arch=mips,mipsel profile=!stage1
 zlib1g deb libs required arch=any
 zlib1g-dbg deb debug extra arch=any
 zlib1g-dev deb libdevel optional arch=any
 zlib1g-udeb udeb debian-installer optional arch=any
Checksums-Sha1:
 1b7f6963ccfb7262a6c9d88894d3a30ff2bf2e23 370248 zlib_1.2.11.dfsg.orig.tar.gz
 c3b2bac9b1927fde66b72d4f98e4063ce0b51f34 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Checksums-Sha256:
 80c481411a4fe8463aeb8270149a0e80bb9eaf7da44132b6e16f2b5af01bc899 370248 zlib_1.2.11.dfsg.orig.tar.gz
 00b95b629fbe9a5181f8ba1ceddedf627aba1ab42e47f5916be8a41deb54098a 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Files:
 2950b229ed4a5e556ad6581580e4ab2c 370248 zlib_1.2.11.dfsg.orig.tar.gz
 fd4b8f37a845569734dfa2e0fe8a08dc 18956 zlib_1.2.11.dfsg-1.debian.tar.xz

-----BEGIN PGP SIGNATURE-----

iQFHBAEBCgAxFiEEreZoqmdXGLWf4p/qJNaLcl1Uh9AFAlrjaY8THGJyb29uaWVA
ZGViaWFuLm9yZwAKCRAk1otyXVSH0J2YB/wMNRqxrwDU5ZJAfnp5gEK1Dq0qT79v
FEEQ6ZfXHnxhQhIQvI/KIPLUIrqkoyvp1DhCxKi+b0LM29JMOpgK8l40EAE52gGk
2vH/NSrcgSXK8yXYJYG0h9fobAihK+eUa6cyp0NR049s8OI00BGSBiHJn3ONH6G4
zNE2vmG2iLVP9bdF/w+0WbyfwccnsjJprrCefJ2/HR+5ckX+PIGklHpI3WVoL1a8
qwfP8hx7pnR4CV1rFfmmDmtwORv4edwfgS6mmbdpKClWS2k4ft7AFEjmBPL+vqKc
6DzauiRhm79p2HZPGWyLvZ0rMUaVvIGYeNjO2n4Lpkc+snZAEX/3LFVQ
=BVVn
-----END PGP SIGNATURE-----
'''

EXPECTED_SIGNED = '''Format: 3.0 (quilt)
Source: zlib
Binary: zlib1g, zlib1g-dev, zlib1g-dbg, zlib1g-udeb, lib64z1, lib64z1-dev, lib32z1, lib32z1-dev, libn32z1, libn32z1-dev
Architecture: any
Version: 1:1.2.11.dfsg-1
Maintainer: Mark Brown <broonie@debian.org>
Homepage: http://zlib.net/
Standards-Version: 3.9.8
Build-Depends: debhelper (>= 8.1.3~), binutils (>= 2.18.1~cvs20080103-2) [mips mipsel], gcc-multilib [amd64 i386 kfreebsd-amd64 mips mipsel powerpc ppc64 s390 sparc s390x] <!stage1>, dpkg-dev (>= 1.16.1)
Package-List:
 lib32z1 deb libs optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib32z1-dev deb libdevel optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib64z1 deb libs optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 lib64z1-dev deb libdevel optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 libn32z1 deb libs optional arch=mips,mipsel profile=!stage1
 libn32z1-dev deb libdevel optional arch=mips,mipsel profile=!stage1
 zlib1g deb libs required arch=any
 zlib1g-dbg deb debug extra arch=any
 zlib1g-dev deb libdevel optional arch=any
 zlib1g-udeb udeb debian-installer optional arch=any
Checksums-Sha1:
 1b7f6963ccfb7262a6c9d88894d3a30ff2bf2e23 370248 zlib_1.2.11.dfsg.orig.tar.gz
 c3b2bac9b1927fde66b72d4f98e4063ce0b51f34 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Checksums-Sha256:
 80c481411a4fe8463aeb8270149a0e80bb9eaf7da44132b6e16f2b5af01bc899 370248 zlib_1.2.11.dfsg.orig.tar.gz
 00b95b629fbe9a5181f8ba1ceddedf627aba1ab42e47f5916be8a41deb54098a 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Files:
 2950b229ed4a5e556ad6581580e4ab2c 370248 zlib_1.2.11.dfsg.orig.tar.gz
 fd4b8f37a845569734dfa2e0fe8a08dc 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
'''

EMPTY = '''-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA512



-----BEGIN PGP SIGNATURE-----

iQFHBAEBCgAxFiEEreZoqmdXGLWf4p/qJNaLcl1Uh9AFAlrjaY8THGJyb29uaWVA
ZGViaWFuLm9yZwAKCRAk1otyXVSH0J2YB/wMNRqxrwDU5ZJAfnp5gEK1Dq0qT79v
FEEQ6ZfXHnxhQhIQvI/KIPLUIrqkoyvp1DhCxKi+b0LM29JMOpgK8l40EAE52gGk
2vH/NSrcgSXK8yXYJYG0h9fobAihK+eUa6cyp0NR049s8OI00BGSBiHJn3ONH6G4
zNE2vmG2iLVP9bdF/w+0WbyfwccnsjJprrCefJ2/HR+5ckX+PIGklHpI3WVoL1a8
qwfP8hx7pnR4CV1rFfmmDmtwORv4edwfgS6mmbdpKClWS2k4ft7AFEjmBPL+vqKc
6DzauiRhm79p2HZPGWyLvZ0rMUaVvIGYeNjO2n4Lpkc+snZAEX/3LFVQ
=BVVn
-----END PGP SIGNATURE-----
'''

COMPACT = '''-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA512
Format: 3.0 (quilt)
Source: zlib
Binary: zlib1g, zlib1g-dev, zlib1g-dbg, zlib1g-udeb, lib64z1, lib64z1-dev, lib32z1, lib32z1-dev, libn32z1, libn32z1-dev
Architecture: any
Version: 1:1.2.11.dfsg-1
Maintainer: Mark Brown <broonie@debian.org>
Homepage: http://zlib.net/
Standards-Version: 3.9.8
Build-Depends: debhelper (>= 8.1.3~), binutils (>= 2.18.1~cvs20080103-2) [mips mipsel], gcc-multilib [amd64 i386 kfreebsd-amd64 mips mipsel powerpc ppc64 s390 sparc s390x] <!stage1>, dpkg-dev (>= 1.16.1)
Package-List:
 lib32z1 deb libs optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib32z1-dev deb libdevel optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib64z1 deb libs optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 lib64z1-dev deb libdevel optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 libn32z1 deb libs optional arch=mips,mipsel profile=!stage1
 libn32z1-dev deb libdevel optional arch=mips,mipsel profile=!stage1
 zlib1g deb libs required arch=any
 zlib1g-dbg deb debug extra arch=any
 zlib1g-dev deb libdevel optional arch=any
 zlib1g-udeb udeb debian-installer optional arch=any
Checksums-Sha1:
 1b7f6963ccfb7262a6c9d88894d3a30ff2bf2e23 370248 zlib_1.2.11.dfsg.orig.tar.gz
 c3b2bac9b1927fde66b72d4f98e4063ce0b51f34 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Checksums-Sha256:
 80c481411a4fe8463aeb8270149a0e80bb9eaf7da44132b6e16f2b5af01bc899 370248 zlib_1.2.11.dfsg.orig.tar.gz
 00b95b629fbe9a5181f8ba1ceddedf627aba1ab42e47f5916be8a41deb54098a 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Files:
 2950b229ed4a5e556ad6581580e4ab2c 370248 zlib_1.2.11.dfsg.orig.tar.gz
 fd4b8f37a845569734dfa2e0fe8a08dc 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
-----BEGIN PGP SIGNATURE-----
iQFHBAEBCgAxFiEEreZoqmdXGLWf4p/qJNaLcl1Uh9AFAlrjaY8THGJyb29uaWVA
ZGViaWFuLm9yZwAKCRAk1otyXVSH0J2YB/wMNRqxrwDU5ZJAfnp5gEK1Dq0qT79v
FEEQ6ZfXHnxhQhIQvI/KIPLUIrqkoyvp1DhCxKi+b0LM29JMOpgK8l40EAE52gGk
2vH/NSrcgSXK8yXYJYG0h9fobAihK+eUa6cyp0NR049s8OI00BGSBiHJn3ONH6G4
zNE2vmG2iLVP9bdF/w+0WbyfwccnsjJprrCefJ2/HR+5ckX+PIGklHpI3WVoL1a8
qwfP8hx7pnR4CV1rFfmmDmtwORv4edwfgS6mmbdpKClWS2k4ft7AFEjmBPL+vqKc
6DzauiRhm79p2HZPGWyLvZ0rMUaVvIGYeNjO2n4Lpkc+snZAEX/3LFVQ
=BVVn
-----END PGP SIGNATURE-----
'''

EXPECTED_COMPACT = '''Hash: SHA512
Format: 3.0 (quilt)
Source: zlib
Binary: zlib1g, zlib1g-dev, zlib1g-dbg, zlib1g-udeb, lib64z1, lib64z1-dev, lib32z1, lib32z1-dev, libn32z1, libn32z1-dev
Architecture: any
Version: 1:1.2.11.dfsg-1
Maintainer: Mark Brown <broonie@debian.org>
Homepage: http://zlib.net/
Standards-Version: 3.9.8
Build-Depends: debhelper (>= 8.1.3~), binutils (>= 2.18.1~cvs20080103-2) [mips mipsel], gcc-multilib [amd64 i386 kfreebsd-amd64 mips mipsel powerpc ppc64 s390 sparc s390x] <!stage1>, dpkg-dev (>= 1.16.1)
Package-List:
 lib32z1 deb libs optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib32z1-dev deb libdevel optional arch=amd64,ppc64,kfreebsd-amd64,s390x profile=!stage1
 lib64z1 deb libs optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 lib64z1-dev deb libdevel optional arch=sparc,s390,i386,powerpc,mips,mipsel profile=!stage1
 libn32z1 deb libs optional arch=mips,mipsel profile=!stage1
 libn32z1-dev deb libdevel optional arch=mips,mipsel profile=!stage1
 zlib1g deb libs required arch=any
 zlib1g-dbg deb debug extra arch=any
 zlib1g-dev deb libdevel optional arch=any
 zlib1g-udeb udeb debian-installer optional arch=any
Checksums-Sha1:
 1b7f6963ccfb7262a6c9d88894d3a30ff2bf2e23 370248 zlib_1.2.11.dfsg.orig.tar.gz
 c3b2bac9b1927fde66b72d4f98e4063ce0b51f34 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Checksums-Sha256:
 80c481411a4fe8463aeb8270149a0e80bb9eaf7da44132b6e16f2b5af01bc899 370248 zlib_1.2.11.dfsg.orig.tar.gz
 00b95b629fbe9a5181f8ba1ceddedf627aba1ab42e47f5916be8a41deb54098a 18956 zlib_1.2.11.dfsg-1.debian.tar.xz
Files:
 2950b229ed4a5e556ad6581580e4ab2c 370248 zlib_1.2.11.dfsg.orig.tar.gz
 fd4b8f37a845569734dfa2e0fe8a08dc 18956 zlib_1.2.11.dfsg-1.debian.tar.xz'''



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#
# Copyright [2017] The Climate Corporation (https://climate.com)
# https://github.com/TheClimateCorporation/python-dpkg
# original author: Nathan J. Meh

from unittest import TestCase

from debian_inspector import version
from debian_inspector.version import compare_strings
from debian_inspector.version import compare_versions
from debian_inspector.version import Version

"""
Parse, compare and sort Debian package versions.

This has been substantially modified from python-dpkg to extract the version
comparison code.
"""


def get_non_digit_prefix(s):
    val = version.get_non_digit_prefix(list(s))
    return ''.join(val)


def get_digit_prefix(s):
    return version.get_digit_prefix(list(s))


class DebianVersionTest(TestCase):

    def test_Version_from_string_epoch(self):
        assert 0 == Version.from_string('0').epoch
        assert 0 == Version.from_string('0:0').epoch
        assert 1 == Version.from_string('1:0').epoch

    def test_Version_from_string_validates(self):
        self.assertRaises(ValueError, Version.from_string, 'a')
        self.assertRaises(ValueError, Version.from_string, '1a:0')
        self.assertRaises(ValueError, Version.from_string, '0:a.0.0-foo')

    def test_Version_from_string_tuples(self):
        assert (0, '00', '0') == Version.from_string('00').tuple()
        assert (0, '00', '00') == Version.from_string('00-00').tuple()
        assert (0, '0', '0') == Version.from_string('0:0').tuple()
        assert (0, '0', '0') == Version.from_string('0:0-0').tuple()
        assert (0, '0.0', '0') == Version.from_string('0:0.0').tuple()
        assert (0, '0.0', '0') == Version.from_string('0:0.0-0').tuple()
        assert (0, '0.0', '00') == Version.from_string('0:0.0-00').tuple()

    def test_get_non_digit_prefix(self):
        assert '' == get_non_digit_prefix('')
        assert '' == get_non_digit_prefix('0')
        assert '' == get_non_digit_prefix('00')
        assert '' == get_non_digit_prefix('0a')
        assert 'a' == get_non_digit_prefix('a')
        assert 'a' == get_non_digit_prefix('a0')
        assert 'aHAD' == get_non_digit_prefix('aHAD0030')

    def test_get_digit_prefix(self):
        assert 0 == get_digit_prefix('00')
        assert 0 == get_digit_prefix('0')
        assert 0 == get_digit_prefix('0a')
        assert 12 == get_digit_prefix('12a')
        assert 0 == get_digit_prefix('a')
        assert 0 == get_digit_prefix('a0')
        assert 0 == get_digit_prefix('arttr23123')
        assert 123 == get_digit_prefix('123sdf')
        assert 123 == get_digit_prefix('0123sdf')

    def test_compare_strings(self):
        assert -1 == compare_strings('~', '.')
        assert -1 == compare_strings('~', 'a')
        assert -1 == compare_strings('a', '.')
        assert 1 == compare_strings('a', '~')
        assert 1 == compare_strings('.', '~')
        assert 1 == compare_strings('.', 'a')
        assert 0 == compare_strings('.', '.')
        assert 0 == compare_strings('0', '0')
        assert 0 == compare_strings('a', 'a')

    def test_compare_strings_can_sort(self):
        # taken from
        # http://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-Version
        result = sorted(['a', '', '~', '~~a', '~~'], key=version.compare_strings_key)
        expected = ['~~', '~~a', '~', '', 'a']
        assert result == expected

    def test_compare_strings_more(self):
        # note that these are testing a single revision string, not the full
        # upstream+debian version.  IOW, "0.0.9-foo" is an upstream or debian
        # revision onto itself, not an upstream of 0.0.9 and a debian of foo.

        # equals
        assert 0 == compare_strings('0', '0')
        assert 0 == compare_strings('0', '00')
        assert 0 == compare_strings('00.0.9', '0.0.9')
        assert 0 == compare_strings('0.00.9-foo', '0.0.9-foo')
        assert 0 == compare_strings('0.0.9-1.00foo', '0.0.9-1.0foo')

        # less than
        assert -1 == compare_strings('0.0.9', '0.0.10')
        assert -1 == compare_strings('0.0.9-foo', '0.0.10-foo')
        assert -1 == compare_strings('0.0.9-foo', '0.0.10-goo')
        assert -1 == compare_strings('0.0.9-foo', '0.0.9-goo')
        assert -1 == compare_strings('0.0.10-foo', '0.0.10-goo')
        assert -1 == compare_strings('0.0.9-1.0foo', '0.0.9-1.1foo')

        # greater than
        assert 1 == compare_strings('0.0.10', '0.0.9')
        assert 1 == compare_strings('0.0.10-foo', '0.0.9-foo')
        assert 1 == compare_strings('0.0.10-foo', '0.0.9-goo')
        assert 1 == compare_strings('0.0.9-1.0foo', '0.0.9-1.0bar')

    def test_compare_versions(self):
        # "This [the epoch] is a single (generally small) unsigned integer.
        # It may be omitted, in which case zero is assumed."
        assert 0 == compare_versions('0.0.0', '0:0.0.0')
        assert 0 == compare_versions('0:0.0.0-foo', '0.0.0-foo')
        assert 0 == compare_versions('0.0.0-a', '0:0.0.0-a')

        # "The absence of a debian_revision is equivalent to a debian_revision of 0."
        assert 0 == compare_versions('0.0.0', '0.0.0-0')
        # tricksy:
        assert 0 == compare_versions('0.0.0', '0.0.0-00')

        # combining the above
        assert 0 == compare_versions('0.0.0-0', '0:0.0.0')

        # explicitly equal
        assert 0 == compare_versions('0.0.0', '0.0.0')
        assert 0 == compare_versions('1:0.0.0', '1:0.0.0')
        assert 0 == compare_versions('0.0.0-10', '0.0.0-10')
        assert 0 == compare_versions('2:0.0.0-1', '2:0.0.0-1')

        # less than
        assert -1 == compare_versions('0.0.0-0', '0:0.0.1')
        assert -1 == compare_versions('0.0.0-0', '0:0.0.0-a')
        assert -1 == compare_versions('0.0.0-0', '0:0.0.0-1')
        assert -1 == compare_versions('0.0.9', '0.0.10')
        assert -1 == compare_versions('0.9.0', '0.10.0')
        assert -1 == compare_versions('9.0.0', '10.0.0')

        assert -1 == compare_versions("1.2.3-1~deb7u1", "1.2.3-1")
        assert -1 == compare_versions("2.7.4+reloaded2-13ubuntu1", "2.7.4+reloaded2-13+deb9u1")
        assert -1 == compare_versions("2.7.4+reloaded2-13", "2.7.4+reloaded2-13+deb9u1")

        # greater than
        assert 1 == compare_versions('0.0.1-0', '0:0.0.0')
        assert 1 == compare_versions('0.0.0-a', '0:0.0.0-1')
        assert 1 == compare_versions('0.0.0-a', '0:0.0.0-0')
        assert 1 == compare_versions('0.0.9', '0.0.1')
        assert 1 == compare_versions('0.9.0', '0.1.0')
        assert 1 == compare_versions('9.0.0', '1.0.0')

        assert 1 == compare_versions("1.2.3-1", "1.2.3-1~deb7u1")
        assert 1 == compare_versions("2.7.4+reloaded2-13+deb9u1", "2.7.4+reloaded2-13ubuntu1")
        assert 1 == compare_versions("2.7.4+reloaded2-13+deb9u1", "2.7.4+reloaded2-13")

        # unicode
        assert -1 == compare_versions(u'2:0.0.44-1', u'2:0.0.44-nobin')
        assert 1 == compare_versions(u'2:0.0.44-nobin', u'2:0.0.44-1')
        assert 0 == compare_versions(u'2:0.0.44-1', u'2:0.0.44-1')



#!/usr/bin/python

# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2005 Florian Weimer <fw@deneb.enyo.de>
# Copyright (C) 2006-2007 James Westby <jw+debian@jameswestby.net>
# Copyright (C) 2010 John Wright <jsw@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# this subset of Python debian tests has been modified to tests version
# comparison and parsing

from unittest import TestCase

from debian_inspector import version


def check_Version_from_string(version_string, epoch, upstream, revision):
    v = version.Version.from_string(version_string)
    result = v.epoch, v.upstream, v.revision
    expected = epoch, upstream, revision
    assert result == expected


def check_compare_versions(version1, operator, version2):
    expected_ops = {
        '==': 0,
        '>': 1,
        '<':-1,
    }
    expected = expected_ops[operator]
    compared = version.compare_versions(version1, version2)
    assert compared == expected


class VersionTests(TestCase):

    def test_Version_from_string(self):
        check_Version_from_string('1:1.4.1-1', 1, '1.4.1', '1')
        check_Version_from_string('7.1.ds-1', 0, '7.1.ds', '1')
        check_Version_from_string('10.11.1.3-2', 0, '10.11.1.3', '2')
        check_Version_from_string('4.0.1.3.dfsg.1-2', 0, '4.0.1.3.dfsg.1', '2')
        check_Version_from_string('0.4.23debian1', 0, '0.4.23debian1', '0')
        check_Version_from_string('1.2.10+cvs20060429-1', 0, '1.2.10+cvs20060429', '1')
        check_Version_from_string('0.2.0-1+b1', 0, '0.2.0', '1+b1')
        check_Version_from_string('4.3.90.1svn-r21976-1', 0, '4.3.90.1svn-r21976', '1')
        check_Version_from_string('1.5+E-14', 0, '1.5+E', '14')
        check_Version_from_string('20060611-0.0', 0, '20060611', '0.0')
        check_Version_from_string('0.52.2-5.1', 0, '0.52.2', '5.1')
        check_Version_from_string('7.0-035+1', 0, '7.0', '035+1')
        check_Version_from_string('1.1.0+cvs20060620-1+2.6.15-8', 0, '1.1.0+cvs20060620-1+2.6.15', '8')
        check_Version_from_string('1.1.0+cvs20060620-1+1.0', 0, '1.1.0+cvs20060620', '1+1.0')
        check_Version_from_string('4.2.0a+stable-2sarge1', 0, '4.2.0a+stable', '2sarge1')
        check_Version_from_string('1.8RC4b', 0, '1.8RC4b', '0')
        check_Version_from_string('0.9~rc1-1', 0, '0.9~rc1', '1')
        check_Version_from_string('2:1.0.4+svn26-1ubuntu1', 2, '1.0.4+svn26', '1ubuntu1')
        check_Version_from_string('2:1.0.4~rc2-1', 2, '1.0.4~rc2', '1')
        self.assertRaises(ValueError, version.Version.from_string, 'a1:1.8.8-070403-1~priv1')

    def test_compare_versions(self):
        check_compare_versions('1.0', '<', '1.1')
        check_compare_versions('1.2', '<', '1.11')
        check_compare_versions('1.0-0.1', '<', '1.1')
        check_compare_versions('1.0-0.1', '<', '1.0-1')
        check_compare_versions('1.0', '==', '1.0')
        check_compare_versions('1.0-0.1', '==', '1.0-0.1')
        check_compare_versions('1:1.0-0.1', '==', '1:1.0-0.1')
        check_compare_versions('1:1.0', '==', '1:1.0')
        check_compare_versions('1.0-0.1', '<', '1.0-1')
        check_compare_versions('1.0final-5sarge1', '>', '1.0final-5')
        check_compare_versions('1.0final-5', '>', '1.0a7-2')
        check_compare_versions('0.9.2-5', '<', '0.9.2+cvs.1.0.dev.2004.07.28-1.5')
        check_compare_versions('1:500', '<', '1:5000')
        check_compare_versions('100:500', '>', '11:5000')
        check_compare_versions('1.0.4-2', '>', '1.0pre7-2')
        check_compare_versions('1.5~rc1', '<', '1.5')
        check_compare_versions('1.5~rc1', '<', '1.5+b1')
        check_compare_versions('1.5~rc1', '<', '1.5~rc2')
        check_compare_versions('1.5~rc1', '>', '1.5~dev0')
